from typing import Callable

class AsyncIOEventEmitter:
    def emit(self, name: str, *args) -> None: ...
    def on(self, name: str, cb: Callable[..., None] = ...) -> Callable[..., None]: ...
    def remove_all_listeners(self) -> None: ...
